@@ -31,7 +31,9 @@
  * Copyright (c) 2003 Per Cederberg. All rights reserved.
  */
 
+using System;
 using System.Collections;
+using System.Text;
 
 namespace PerCederberg.Grammatica.Parser {
 
@@ -40,39 +42,59 @@ namespace PerCederberg.Grammatica.Parser {
      * interface, as well as token handling.
      *
      * @author   Per Cederberg, <per at percederberg dot net>
-     * @version  1.0
+     * @version  1.1
      */
     public abstract class Parser {
-
+    
+        /**
+         * The parser initialization flag.
+         */
+        private bool initialized = false;
+    
         /**
          * The tokenizer to use.
          */
         private Tokenizer tokenizer;
-
+    
         /**
          * The analyzer to use for callbacks.
          */
         private Analyzer analyzer;
-
+    
         /**
-         * The start production pattern id. This will be set to the
-         * first production pattern added.
+         * The list of production patterns. 
          */
-        private int start = 0;
-
+        private ArrayList patterns = new ArrayList();
+    
         /**
-         * The map with all production patterns. The production
-         * patterns are indexed by their id.
+         * The map with production patterns and their id:s. This map 
+         * contains the production patterns indexed by their id:s.
          */
-        private Hashtable patterns = new Hashtable();
-
+        private Hashtable patternIds = new Hashtable();
+    
         /**
-         * The list of buffered tokens. This list will contain tokens
-         * that have been read from the tokenizer, but not yet
-         * consumed.
+         * The list of buffered tokens. This list will contain tokens that
+         * have been read from the tokenizer, but not yet consumed.
          */
         private ArrayList tokens = new ArrayList();
-
+        
+        /**
+         * The error log. All parse errors will be added to this log as
+         * the parser attempts to recover from the error. If the error
+         * count is higher than zero (0), this log will be thrown as the
+         * result from the parse() method. 
+         */
+        private ParserLogException errorLog = new ParserLogException();
+    
+        /**
+         * The error recovery counter. This counter is initially set to a
+         * negative value to indicate that no error requiring recovery 
+         * has been encountered. When a parse error is found, the counter
+         * is set to three (3), and is then decreased by one for each 
+         * correctly read token until it reaches zero (0).  
+         */
+        private int errorRecovery = -1;
+    
         /**
          * Creates a new parser.
          * 
@@ -93,6 +115,17 @@ namespace PerCederberg.Grammatica.Parser {
             this.analyzer = analyzer;
         }
 
+        /**
+         * Sets the parser initialized flag. Normally this flag is set by
+         * the prepare() method, but this method allows further 
+         * modifications to it.
+         * 
+         * @param initialized    the new initialized flag
+         */
+        internal void SetInitialized(bool initialized) {
+            this.initialized = initialized;
+        }
+    
         /**
          * Adds a new production pattern to the parser. The first pattern 
          * added is assumed to be the starting point in the grammar. The 
@@ -111,42 +144,38 @@ namespace PerCederberg.Grammatica.Parser {
                     "no production alternatives are present (must have at " +
                     "least one)");
             }
-            if (patterns.Count == 0) {
-                start = pattern.GetId();
-            }
-            if (patterns.ContainsKey(pattern.GetId())) {
+            if (patternIds.ContainsKey(pattern.GetId())) {
                 throw new ParserCreationException(
                     ParserCreationException.ErrorType.INVALID_PRODUCTION,
                     pattern.GetName(),
                     "another pattern with the same id (" + pattern.GetId() + 
                     ") has already been added");
             }
-            patterns.Add(pattern.GetId(), pattern);
+            patterns.Add(pattern);
+            patternIds.Add(pattern.GetId(), pattern);
+            SetInitialized(false);
         }
-
+    
         /**
-         * Initializes the parser. All the added production patterns
-         * will be analyzed for ambiguities and errors. This method
-         * also initializes internal data structures used during the
-         * parsing.
+         * Initializes the parser. All the added production patterns will
+         * be analyzed for ambiguities and errors. This method also 
+         * initializes internal data structures used during the parsing. 
          * 
          * @throws ParserCreationException if the parser couldn't be 
          *             initialized correctly 
          */
         public virtual void Prepare() {
-            IDictionaryEnumerator  e;
-            
-            if (patterns.Count == 0) {
+            if (patterns.Count <= 0) {
                 throw new ParserCreationException(
                     ParserCreationException.ErrorType.INVALID_PARSER,
                     "no production patterns have been added");
             }
-            e = patterns.GetEnumerator();
-            while (e.MoveNext()) {
-                CheckPattern((ProductionPattern) e.Value);
+            for (int i = 0; i < patterns.Count; i++) {
+                CheckPattern((ProductionPattern) patterns[i]);
             }
+            SetInitialized(true);
         }
-
+    
         /**
          * Checks a production pattern for completeness. If some rule
          * in the pattern referenced an production pattern not added
@@ -208,19 +237,72 @@ namespace PerCederberg.Grammatica.Parser {
 
         /**
          * Parses the token stream and returns a parse tree. This method
-         * will call prepare() if not previously called.
+         * will call prepare() if not previously called. In case of a 
+         * parse error, the parser will attempt to recover and throw all
+         * the errors found in a parser log exception in the end.
          * 
          * @return the parse tree
          * 
          * @throws ParserCreationException if the parser couldn't be
          *             initialized correctly
-         * @throws ParseException if the input couldn't be parsed 
+         * @throws ParserLogException if the input couldn't be parsed 
          *             correctly
          * 
          * @see #prepare
          */
-        public abstract Node Parse();
-
+        public Node Parse() {
+            Node  root = null;
+    
+            // Initialize parser
+            if (!initialized) {
+                Prepare();
+            }
+    
+            // Parse input
+            try {
+                root = ParseStart();
+            } catch (ParseException e) {
+                AddError(e, true);
+            }
+            
+            // Check for errors
+            if (errorLog.GetErrorCount() > 0) {
+                throw errorLog;
+            }
+    
+            return root;
+        }
+    
+        /**
+         * Parses the token stream and returns a parse tree.
+         * 
+         * @return the parse tree
+         * 
+         * @throws ParseException if the input couldn't be parsed 
+         *             correctly
+         */
+        protected abstract Node ParseStart();
+    
+        /**
+         * Adds an error to the error log. If the parser is in error 
+         * recovery mode, the error will not be added to the log. If the
+         * recovery flag is set, this method will set the error recovery 
+         * counter thus enter error recovery mode. Only lexical or 
+         * syntactical errors require recovery, so this flag shouldn't be
+         * set otherwise.
+         * 
+         * @param e              the error to add
+         * @param recovery       the recover flag 
+         */
+        internal void AddError(ParseException e, bool recovery) {
+            if (errorRecovery <= 0) {
+                errorLog.AddError(e);
+            }
+            if (recovery) {
+                errorRecovery = 3;
+            }
+        }
+    
         /**
          * Returns the production pattern with the specified id.
          *  
@@ -230,7 +312,7 @@ namespace PerCederberg.Grammatica.Parser {
          *         null if non-existent
          */
         internal ProductionPattern GetPattern(int id) {
-            return (ProductionPattern) patterns[id];
+            return (ProductionPattern) patternIds[id];
         }
 
         /**
@@ -240,77 +322,89 @@ namespace PerCederberg.Grammatica.Parser {
          *         null if no patterns have been added
          */
         internal ProductionPattern GetStartPattern() {
-            return GetPattern(start);
-        }
-
-        /**
-         * Returns the unordered set of production patterns.
-         * 
-         * @return the unordered set of production patterns
-         */
-        internal ICollection GetPatterns() {
-            return patterns.Values;
-        }
-
-        /**
-         * Handles the parser entering a production. This method calls
-         * the appropriate analyzer callback if the node is not
-         * hidden.
-         * 
-         * @param node            the parse tree node
-         * 
-         * @throws ParseException if the node couldn't be handled 
-         *             correctly
-         */
-        internal void EnterNode(Node node) {
-            if (!node.IsHidden()) {
-                analyzer.Enter(node);
+            if (patterns.Count <= 0) {
+                return null;
+            } else {
+                return (ProductionPattern) patterns[0];
             }
         }
     
         /**
-         * Handles the parser leaving a production. This method calls
-         * the appropriate analyzer callback if the node is not
-         * hidden, and returns the result.
+         * Returns the ordered set of production patterns.
+         * 
+         * @return the ordered set of production patterns
+         */
+        internal ICollection GetPatterns() {
+            return patterns;
+        }
+
+        /**
+         * Handles the parser entering a production. This method calls the
+         * appropriate analyzer callback if the node is not hidden. Note
+         * that this method will not call any callback if an error 
+         * requiring recovery has ocurred.
+         * 
+         * @param node           the parse tree node
+         */
+        internal void EnterNode(Node node) {
+            if (!node.IsHidden() && errorRecovery < 0) {
+                try {
+                    analyzer.Enter(node);
+                } catch (ParseException e) {
+                    AddError(e, false);
+                }
+            }
+        }
+        
+        /**
+         * Handles the parser leaving a production. This method calls the
+         * appropriate analyzer callback if the node is not hidden, and 
+         * returns the result. Note that this method will not call any 
+         * callback if an error requiring recovery has ocurred.
          * 
          * @param node           the parse tree node
          * 
          * @return the parse tree node, or
          *         null if no parse tree should be created
-         * 
-         * @throws ParseException if the node couldn't be handled 
-         *             correctly
          */
         internal Node ExitNode(Node node) {
-            if (node.IsHidden()) {
-                return node;
-            } else {
-                return analyzer.Exit(node);
+            if (!node.IsHidden() && errorRecovery < 0) {
+                try {
+                    return analyzer.Exit(node);
+                } catch (ParseException e) {
+                    AddError(e, false);
+                }
             }
+            return node;
         }
-
+    
         /**
-         * Handles the parser adding a child node to a production.
-         * This method calls the appropriate analyzer callback.
+         * Handles the parser adding a child node to a production. This 
+         * method calls the appropriate analyzer callback. Note that this 
+         * method will not call any callback if an error requiring 
+         * recovery has ocurred.
          * 
          * @param node           the parent parse tree node
          * @param child          the child parse tree node, or null
-         * 
-         * @throws ParseException if the node couldn't be handled 
-         *             correctly
          */
         internal void AddNode(Production node, Node child) {
-            if (node.IsHidden()) {
+            if (errorRecovery >= 0) {
+                // Do nothing
+            } else if (node.IsHidden()) {
                 node.AddChild(child);
             } else if (child != null && child.IsHidden()) {
                 for (int i = 0; i < child.GetChildCount(); i++) {
                     AddNode(node, child.GetChildAt(i));
                 }
             } else {
-                analyzer.Child(node, child);
+                try {
+                    analyzer.Child(node, child);
+                } catch (ParseException e) {
+                    AddError(e, false);
+                }
             }
         }
-
+    
         /**
          * Reads and consumes the next token in the queue. If no token
          * was available for consumation, a parse error will be
@@ -354,6 +448,9 @@ namespace PerCederberg.Grammatica.Parser {
             ArrayList  list;
             
             if (token.GetId() == id) {
+                if (errorRecovery > 0) {
+                    errorRecovery--;
+                }
                 return token;
             } else {
                 list = new ArrayList(1);
@@ -366,34 +463,153 @@ namespace PerCederberg.Grammatica.Parser {
                     token.GetStartColumn());
             }
         }
-
+    
         /**
-         * Returns a token from the queue. This method is used to
-         * check coming tokens before they have been consumed. Any
-         * number of tokens forward can be checked.
+         * Returns a token from the queue. This method is used to check 
+         * coming tokens before they have been consumed. Any number of 
+         * tokens forward can be checked. 
          * 
          * @param steps          the token queue number, zero (0) for first
          * 
          * @return the token in the queue, or
          *         null if no more tokens in the queue
-         * 
-         * @throws ParseException if the input stream couldn't be read or
-         *             parsed correctly
          */
         internal Token PeekToken(int steps) {
             Token  token;
-
+    
             while (steps >= tokens.Count) {
-                token = tokenizer.Next();
-                if (token == null) {
-                    return null;
-                } else {
-                    tokens.Add(token);
+                try {
+                    token = tokenizer.Next();
+                    if (token == null) {
+                        return null;
+                    } else {
+                        tokens.Add(token);
+                    }
+                } catch (ParseException e) {
+                    AddError(e, true);
                 }
             }
             return (Token) tokens[steps];
         }
-
+    
+        /**
+         * Returns a string representation of this parser. The string will
+         * contain all the production definitions and various additional 
+         * information.
+         * 
+         * @return a detailed string representation of this parser
+         */
+        public override string ToString() {
+            StringBuilder  buffer = new StringBuilder();
+            
+            for (int i = 0; i < patterns.Count; i++) {
+                buffer.Append(ToString((ProductionPattern) patterns[i])); 
+                buffer.Append("\n");
+            }
+            return buffer.ToString();
+        }
+    
+        /**
+         * Returns a string representation of a production pattern.
+         * 
+         * @param prod           the production pattern
+         * 
+         * @return a detailed string representation of the pattern
+         */
+        private string ToString(ProductionPattern prod) {
+            StringBuilder  buffer = new StringBuilder();
+            StringBuilder  indent = new StringBuilder();
+            LookAheadSet   set;
+            int            i;
+    
+            buffer.Append(prod.GetName());
+            buffer.Append(" (");
+            buffer.Append(prod.GetId());
+            buffer.Append(") ");
+            for (i = 0; i < buffer.Length; i++) {
+                indent.Append(" ");
+            }
+            buffer.Append("= ");
+            indent.Append("| ");
+            for (i = 0; i < prod.GetAlternativeCount(); i++) {
+                if (i > 0) {
+                    buffer.Append(indent);
+                }
+                buffer.Append(ToString(prod.GetAlternative(i)));
+                buffer.Append("\n");
+            }
+            for (i = 0; i < prod.GetAlternativeCount(); i++) {
+                set = prod.GetAlternative(i).GetLookAhead();
+                if (set.GetMaxLength() > 1) {
+                    buffer.Append("Using ");
+                    buffer.Append(set.GetMaxLength());
+                    buffer.Append(" token look-ahead for alternative "); 
+                    buffer.Append(i + 1);
+                    buffer.Append(": ");
+                    buffer.Append(set.ToString(tokenizer));
+                    buffer.Append("\n");
+                }
+            }
+            return buffer.ToString();
+        }
+        
+        /**
+         * Returns a string representation of a production pattern 
+         * alternative.
+         * 
+         * @param alt            the production pattern alternative
+         * 
+         * @return a detailed string representation of the alternative
+         */
+        private string ToString(ProductionPatternAlternative alt) {
+            StringBuilder  buffer = new StringBuilder();
+    
+            for (int i = 0; i < alt.GetElementCount(); i++) {
+                if (i > 0) {
+                    buffer.Append(" ");
+                }
+                buffer.Append(ToString(alt.GetElement(i)));
+            }
+            return buffer.ToString();
+        }
+    
+        /**
+         * Returns a string representation of a production pattern 
+         * element.
+         * 
+         * @param elem           the production pattern element
+         * 
+         * @return a detailed string representation of the element
+         */
+        private string ToString(ProductionPatternElement elem) {
+            StringBuilder  buffer = new StringBuilder();
+            int            min = elem.GetMinCount();
+            int            max = elem.GetMaxCount();
+    
+            if (min == 0 && max == 1) {
+                buffer.Append("[");
+            }
+            if (elem.IsToken()) {
+                buffer.Append(GetTokenDescription(elem.GetId()));
+            } else {
+                buffer.Append(GetPattern(elem.GetId()).GetName());
+            }
+            if (min == 0 && max == 1) {
+                buffer.Append("]");
+            } else if (min == 0 && max == Int32.MaxValue) {
+                buffer.Append("*");
+            } else if (min == 1 && max == Int32.MaxValue) {
+                buffer.Append("+");
+            } else if (min != 1 || max != 1) {
+                buffer.Append("{");
+                buffer.Append(min);
+                buffer.Append(",");
+                buffer.Append(max);
+                buffer.Append("}");
+            }
+            return buffer.ToString();
+        }
+    
         /**
          * Returns a token description for a specified token.
          * 
