@@ -47,11 +47,6 @@ namespace PerCederberg.Grammatica.Parser {
      */
     public class RecursiveDescentParser : Parser {
 
-        /**
-         * The parser initialization flag.
-         */
-        private bool initialized = false;
-
         /**
          * The map of pattern look-ahead sets. The map is indexed by
          * the production pattern object.
@@ -110,7 +105,6 @@ namespace PerCederberg.Grammatica.Parser {
 
             // Add pattern 
             base.AddPattern(pattern);
-            initialized = false;
         }
 
         /**
@@ -127,6 +121,7 @@ namespace PerCederberg.Grammatica.Parser {
             
             // Performs production pattern checks
             base.Prepare();
+	        SetInitialized(false);
             
             // Calculate production look-ahead sets
             e = GetPatterns().GetEnumerator();
@@ -135,7 +130,7 @@ namespace PerCederberg.Grammatica.Parser {
             }
             
             // Set initialized flag
-            initialized = true;
+	        SetInitialized(true);
         }
 
         /**
@@ -143,21 +138,14 @@ namespace PerCederberg.Grammatica.Parser {
          * 
          * @return the parse tree
          * 
-         * @throws ParserCreationException if the parser couldn't be
-         *             initialized correctly
          * @throws ParseException if the input couldn't be parsed 
          *             correctly
-         * 
-         * @see #prepare
          */
-        public override Node Parse() {
+        protected override Node ParseStart() {
             Token      token;
             Node       node;
             ArrayList  list;
         
-            if (!initialized) {
-                Prepare();
-            }
             node = ParsePattern(GetStartPattern());
             token = PeekToken(0); 
             if (token != null) {
@@ -220,7 +208,13 @@ namespace PerCederberg.Grammatica.Parser {
             node = new Production(alt.GetPattern());
             EnterNode(node);
             for (int i = 0; i < alt.GetElementCount(); i++) {
-                ParseElement(node, alt.GetElement(i));
+                try {
+                    ParseElement(node, alt.GetElement(i));
+                } catch (ParseException e) {
+                    AddError(e, true);
+                    NextToken();
+                    i--;
+                }
             }
             return ExitNode(node);
         }
@@ -266,9 +260,6 @@ namespace PerCederberg.Grammatica.Parser {
          * 
          * @return true if the next tokens match, or
          *         false otherwise
-         * 
-         * @throws ParseException if a token couldn't be read or parsed 
-         *             correctly
          */
         private bool IsNext(ProductionPattern pattern) {
             LookAheadSet  set = pattern.GetLookAhead();
@@ -289,9 +280,6 @@ namespace PerCederberg.Grammatica.Parser {
          *  
          * @return true if the next tokens match, or
          *         false otherwise
-         * 
-         * @throws ParseException if a token couldn't be read or parsed 
-         *             correctly
          */
         private bool IsNext(ProductionPatternAlternative alt) {
             LookAheadSet  set = alt.GetLookAhead();
@@ -313,9 +301,6 @@ namespace PerCederberg.Grammatica.Parser {
          *  
          * @return true if the next tokens match, or
          *         false otherwise
-         * 
-         * @throws ParseException if a token couldn't be read or parsed 
-         *             correctly
          */
         private bool IsNext(ProductionPatternElement elem) {
             LookAheadSet  set = elem.GetLookAhead();
@@ -330,10 +315,9 @@ namespace PerCederberg.Grammatica.Parser {
         }
 
         /**
-         * Calculates the look-ahead needed for the specified
-         * production pattern. This method attempts to resolve any
-         * conflicts and stores the results in the pattern look-ahead
-         * object.
+         * Calculates the look-ahead needed for the specified production 
+         * pattern. This method attempts to resolve any conflicts and 
+         * stores the results in the pattern look-ahead object. 
          * 
          * @param pattern        the production pattern
          * 
@@ -345,36 +329,40 @@ namespace PerCederberg.Grammatica.Parser {
             LookAheadSet                  result;
             LookAheadSet[]                alternatives;
             LookAheadSet                  conflicts;
+            LookAheadSet                  previous = new LookAheadSet(0);
             int                           length = 1;
             int                           i;
             CallStack                     stack = new CallStack();
-        
+            
             // Calculate simple look-ahead
             stack.Push(pattern.GetName(), 1);
             result = new LookAheadSet(1);
             alternatives = new LookAheadSet[pattern.GetAlternativeCount()];
             for (i = 0; i < pattern.GetAlternativeCount(); i++) {
-                alternatives[i] = FindLookAhead(pattern.GetAlternative(i), 
-                                                1,
-                                                stack);
+                alt = pattern.GetAlternative(i);
+                alternatives[i] = FindLookAhead(alt, 1, 0, stack, null);
+                alt.SetLookAhead(alternatives[i]);
                 result.AddAll(alternatives[i]);
             }
             if (pattern.GetLookAhead() == null) {
                 pattern.SetLookAhead(result);
             }
             conflicts = FindConflicts(pattern, 1);
-
+    
             // Resolve conflicts
             while (conflicts.Size() > 0) {
                 length++;
                 stack.Clear();
                 stack.Push(pattern.GetName(), length);
+                conflicts.AddAll(previous);
                 for (i = 0; i < pattern.GetAlternativeCount(); i++) {
                     alt = pattern.GetAlternative(i);
                     if (alternatives[i].Intersects(conflicts)) {
                         alternatives[i] = FindLookAhead(alt, 
                                                         length,
-                                                        stack);
+                                                        0,
+                                                        stack,
+                                                        conflicts);
                         alt.SetLookAhead(alternatives[i]);
                     }
                     if (alternatives[i].Intersects(conflicts)) {
@@ -388,6 +376,7 @@ namespace PerCederberg.Grammatica.Parser {
                         }
                     }
                 }
+                previous = conflicts;
                 conflicts = FindConflicts(pattern, length);
             }
             
@@ -396,12 +385,12 @@ namespace PerCederberg.Grammatica.Parser {
                 CalculateLookAhead(pattern.GetAlternative(i), 0);
             }
         }
-
+    
         /**
-         * Calculates the look-aheads needed for the specified pattern
-         * alternative. This method attempts to resolve any conflicts
-         * in optional elements by recalculating look-aheads for
-         * referenced productions.
+         * Calculates the look-aheads needed for the specified pattern 
+         * alternative. This method attempts to resolve any conflicts in 
+         * optional elements by recalculating look-aheads for referenced 
+         * productions. 
          * 
          * @param alt            the production pattern alternative
          * @param pos            the pattern element position
@@ -411,20 +400,21 @@ namespace PerCederberg.Grammatica.Parser {
          */
         private void CalculateLookAhead(ProductionPatternAlternative alt,
                                         int pos) {
-
+    
             ProductionPattern         pattern;
             ProductionPatternElement  elem;
             LookAheadSet              first;
             LookAheadSet              follow;
             LookAheadSet              conflicts;
-            string                    location;
+            LookAheadSet              previous = new LookAheadSet(0);
+            String                    location;
             int                       length = 1;
-
+    
             // Check trivial cases
             if (pos >= alt.GetElementCount()) {
                 return;
             }
-
+    
             // Check for non-optional element
             pattern = alt.GetPattern();
             elem = alt.GetElement(pos);
@@ -432,11 +422,11 @@ namespace PerCederberg.Grammatica.Parser {
                 CalculateLookAhead(alt, pos + 1);
                 return;
             }
-
+    
             // Calculate simple look-aheads
-            first = FindLookAhead(elem, 1, new CallStack());
-            follow = FindLookAhead(alt, 1, pos + 1, new CallStack()); 
-        
+            first = FindLookAhead(elem, 1, new CallStack(), null);
+            follow = FindLookAhead(alt, 1, pos + 1, new CallStack(), null); 
+            
             // Resolve conflicts
             location = "at position " + (pos + 1);
             conflicts = FindConflicts(pattern.GetName(), 
@@ -445,33 +435,43 @@ namespace PerCederberg.Grammatica.Parser {
                                       follow);
             while (conflicts.Size() > 0) {
                 length++;
-                first = FindLookAhead(elem, length, new CallStack());
-                follow = FindLookAhead(alt, length, pos + 1, new CallStack());
+                conflicts.AddAll(previous);
+                first = FindLookAhead(elem, 
+                                      length, 
+                                      new CallStack(), 
+                                      conflicts);
+                follow = FindLookAhead(alt, 
+                                       length, 
+                                       pos + 1, 
+                                       new CallStack(), 
+                                       conflicts);
                 first = first.CreateCombination(follow);
                 elem.SetLookAhead(first);
                 if (first.Intersects(conflicts)) {
                     first = first.CreateIntersection(conflicts);
                     ThrowAmbiguityException(pattern.GetName(), location, first);
                 }
+                previous = conflicts;
                 conflicts = FindConflicts(pattern.GetName(), 
                                           location, 
                                           first, 
                                           follow);
             }
-            
+    
             // Check remaining elements
             CalculateLookAhead(alt, pos + 1);
         }
-
+    
         /**
-         * Finds the look-ahead set for a production pattern. The
-         * maximum look-ahead length must be specified. This method
-         * will cache some results in the pattern look-ahead wrapper
-         * for improved performance.
+         * Finds the look-ahead set for a production pattern. The maximum 
+         * look-ahead length must be specified. It is also possible to 
+         * specify a look-ahead set filter, which will make sure that 
+         * unnecessary token sequences will be avoided.
          * 
          * @param pattern        the production pattern
          * @param length         the maximum look-ahead length
          * @param stack          the call stack used for loop detection
+         * @param filter         the look-ahead set filter
          * 
          * @return the look-ahead set for the production pattern
          * 
@@ -480,94 +480,48 @@ namespace PerCederberg.Grammatica.Parser {
          */
         private LookAheadSet FindLookAhead(ProductionPattern pattern, 
                                            int length,
-                                           CallStack stack) {
-
+                                           CallStack stack,
+                                           LookAheadSet filter) {
+    
             LookAheadSet  result;
             LookAheadSet  temp;
-
+    
             // Check for infinite loop
             if (stack.Contains(pattern.GetName(), length)) {
                 throw new ParserCreationException(
                     ParserCreationException.ErrorType.INFINITE_LOOP,
                     pattern.GetName(),
-                    (string) null);
+                    (String) null);
             }
-
-            // Check for cached result
-            result = pattern.GetLookAhead();
-            if (result != null) {
-                if (length == result.GetMaxLength()) {
-                    return result;
-                } else if (length < result.GetMaxLength()) {
-                    return new LookAheadSet(length, result);
-                }
-            }
-
+    
             // Find pattern look-ahead
             stack.Push(pattern.GetName(), length);
             result = new LookAheadSet(length);
             for (int i = 0; i < pattern.GetAlternativeCount(); i++) {
-                temp = FindLookAhead(pattern.GetAlternative(i), length, stack);
+                temp = FindLookAhead(pattern.GetAlternative(i), 
+                                     length,
+                                     0,
+                                     stack,
+                                     filter);
                 result.AddAll(temp);
             }
-            if (pattern.GetLookAhead() == null) {
-                pattern.SetLookAhead(result);
-            }
             stack.Pop();
-
+    
             return result;
         }
-
+    
         /**
-         * Finds the look-ahead set for a production pattern
-         * alternative. The maximum look-ahead length must be
-         * specified. This method will cache some results in the
-         * pattern look-ahead wrapper for improved performance.
-         * 
-         * @param alt            the production pattern alternative
-         * @param length         the maximum look-ahead length
-         * @param stack          the call stack used for loop detection
-         * 
-         * @return the look-ahead set for the pattern alternative
-         * 
-         * @throws ParserCreationException if an infinite loop was found
-         *             in the grammar
-         */
-        private LookAheadSet FindLookAhead(ProductionPatternAlternative alt, 
-                                           int length,
-                                           CallStack stack) {
-
-            LookAheadSet  result;
-                                
-            // Check for cached result
-            result = alt.GetLookAhead();
-            if (result != null) {
-                if (length == result.GetMaxLength()) {
-                    return result;
-                } else if (length < result.GetMaxLength()) {
-                    return new LookAheadSet(length, result); 
-                }
-            }
-
-            // Find rule look-ahead
-            result = FindLookAhead(alt, length, 0, stack);
-            if (alt.GetLookAhead() == null) {
-                alt.SetLookAhead(result);
-            }
-            
-            return result;
-        }
-
-
-        /**
-         * Finds the look-ahead set for a production pattern
-         * alternative. The pattern position and maximum look-ahead
-         * length must be specified.
+         * Finds the look-ahead set for a production pattern alternative. 
+         * The pattern position and maximum look-ahead length must be 
+         * specified. It is also possible to specify a look-ahead set 
+         * filter, which will make sure that unnecessary token sequences 
+         * will be avoided.
          * 
          * @param alt            the production pattern alternative
          * @param length         the maximum look-ahead length
          * @param pos            the pattern element position
          * @param stack          the call stack used for loop detection
+         * @param filter         the look-ahead set filter
          * 
          * @return the look-ahead set for the pattern alternative
          * 
@@ -577,42 +531,56 @@ namespace PerCederberg.Grammatica.Parser {
         private LookAheadSet FindLookAhead(ProductionPatternAlternative alt,
                                            int length,
                                            int pos,
-                                           CallStack stack) {
-
+                                           CallStack stack,
+                                           LookAheadSet filter) {
+    
             LookAheadSet  first;
             LookAheadSet  follow;
-
+            LookAheadSet  overlaps;
+    
             // Check trivial cases
             if (length <= 0 || pos >= alt.GetElementCount()) {
                 return new LookAheadSet(0);
             }
-
+    
             // Find look-ahead for this element
-            first = FindLookAhead(alt.GetElement(pos), length, stack);
+            first = FindLookAhead(alt.GetElement(pos), length, stack, filter);
             if (alt.GetElement(pos).GetMinCount() == 0) {
                 first.AddEmpty();
             }
-        
+            
             // Find remaining look-ahead
-            length -= first.GetMinLength();
-            if (length <= 0) {
-                return first;
-            } else {
-                follow = FindLookAhead(alt, length, pos + 1, stack); 
-                return first.CreateCombination(follow); 
+            if (filter == null) {
+                length -= first.GetMinLength();
+                if (length > 0) {
+                    follow = FindLookAhead(alt, length, pos + 1, stack, null); 
+                    first = first.CreateCombination(follow);
+                }
+            } else if (filter.IsOverlap(first)) {
+                overlaps = first.CreateOverlaps(filter);
+                length -= overlaps.GetMinLength();
+                filter = filter.CreateFilter(overlaps);
+                follow = FindLookAhead(alt, length, pos + 1, stack, filter); 
+                first.RemoveAll(overlaps);
+                first.AddAll(overlaps.CreateCombination(follow));
             }
+    
+            return first;
         }
-
+    
         /**
-         * Finds the look-ahead set for a production pattern element.
-         * The maximum look-ahead length must be specified. This
-         * method takes the element repeats into consideration when
-         * creating the look-ahead set, but does NOT include an empty
-         * sequence even if the minimum count is zero (0).
+         * Finds the look-ahead set for a production pattern element. The
+         * maximum look-ahead length must be specified. This method takes
+         * the element repeats into consideration when creating the 
+         * look-ahead set, but does NOT include an empty sequence even if
+         * the minimum count is zero (0). It is also possible to specify a
+         * look-ahead set filter, which will make sure that unnecessary 
+         * token sequences will be avoided.
          * 
          * @param elem           the production pattern element
          * @param length         the maximum look-ahead length
          * @param stack          the call stack used for loop detection
+         * @param filter         the look-ahead set filter
          * 
          * @return the look-ahead set for the pattern element
          * 
@@ -621,43 +589,90 @@ namespace PerCederberg.Grammatica.Parser {
          */
         private LookAheadSet FindLookAhead(ProductionPatternElement elem,
                                            int length,
-                                           CallStack stack) {
-
+                                           CallStack stack,
+                                           LookAheadSet filter) {
+    
             ProductionPattern  pattern;
             LookAheadSet       result;
-            LookAheadSet       set;
+            LookAheadSet       first;
+            LookAheadSet       follow;
             int                max;
-
-            // Find look-ahead for this element
-            if (elem.IsToken()) {
-                set = new LookAheadSet(length);
-                set.Add(elem.GetId());
-            } else {
-                pattern = GetPattern(elem.GetId());
-                set = FindLookAhead(pattern, length, stack);
-                if (stack.Contains(pattern.GetName())) {
-                    set = set.CreateRepetitive();
-                }
-            }
-
-            // Handle element repetitions
+    
+            // Find initial element look-ahead
+            first = FindLookAhead(elem, length, 0, stack, filter);
             result = new LookAheadSet(length);
-            result.AddAll(set);
+            result.AddAll(first);
+            if (filter == null || !filter.IsOverlap(result)) {
+                return result;
+            }
+    
+            // Handle element repetitions
             if (elem.GetMaxCount() == Int32.MaxValue) {
-                set = set.CreateRepetitive();
+                first = first.CreateRepetitive();
             }
             max = elem.GetMaxCount();
             if (length < max) {
                 max = length;
             }
             for (int i = 1; i < max; i++) {
-                set = set.CreateCombination(set);
-                result.AddAll(set);
+                first = first.CreateOverlaps(filter);
+                if (first.Size() <= 0 || first.GetMinLength() >= length) {
+                    break;
+                }
+                follow = FindLookAhead(elem, 
+                                       length, 
+                                       0, 
+                                       stack, 
+                                       filter.CreateFilter(first));
+                first = first.CreateCombination(follow);
+                result.AddAll(first);
             }
-            
+    
             return result;
         }
-
+    
+        /**
+         * Finds the look-ahead set for a production pattern element. The
+         * maximum look-ahead length must be specified. This method does
+         * NOT take the element repeat into consideration when creating 
+         * the look-ahead set. It is also possible to specify a look-ahead
+         * set filter, which will make sure that unnecessary token 
+         * sequences will be avoided.
+         * 
+         * @param elem           the production pattern element
+         * @param length         the maximum look-ahead length
+         * @param dummy          a parameter to distinguish the method
+         * @param stack          the call stack used for loop detection
+         * @param filter         the look-ahead set filter
+         * 
+         * @return the look-ahead set for the pattern element
+         * 
+         * @throws ParserCreationException if an infinite loop was found
+         *             in the grammar
+         */
+        private LookAheadSet FindLookAhead(ProductionPatternElement elem,
+                                           int length,
+                                           int dummy,
+                                           CallStack stack,
+                                           LookAheadSet filter) {
+    
+            LookAheadSet       result;
+            ProductionPattern  pattern;
+    
+            if (elem.IsToken()) {
+                result = new LookAheadSet(length);
+                result.Add(elem.GetId());
+            } else {
+                pattern = GetPattern(elem.GetId());
+                result = FindLookAhead(pattern, length, stack, filter);
+                if (stack.Contains(pattern.GetName())) {
+                    result = result.CreateRepetitive();
+                }
+            }
+    
+            return result;
+        }
+    
         /**
          * Returns a look-ahead set with all conflics between
          * alternatives in a production pattern.
