@@ -46,7 +46,7 @@ namespace PerCederberg.Grammatica.Parser {
      * don't match any of the token patterns, a parse exception is thrown. 
      *
      * @author   Per Cederberg, <per at percederberg dot net>
-     * @version  1.0
+     * @version  1.1
      */
     public class Tokenizer {
     
@@ -223,9 +223,10 @@ namespace PerCederberg.Grammatica.Parser {
          *             parsed correctly
          */
         private Token NextToken() {
-            TokenMatcher  m;
-            Token         token;
-            string        str;
+            TokenMatcher   m;
+            Token          token;
+            string         str;
+           ParseException  e;
             
             // Find longest matching string 
             do {
@@ -248,11 +249,19 @@ namespace PerCederberg.Grammatica.Parser {
             } else if (position >= buffer.Length) {
                 return null;
             } else {
-                throw new ParseException(
+                e = new ParseException(
                     ParseException.ErrorType.UNEXPECTED_CHAR,
                     buffer[position].ToString(),
                     line,
-                    column);
+                    column); 
+                if (buffer[position] == '\n') {
+                    line++;
+                    column = 1;
+                } else {
+                    column++;
+                }
+                position++;
+                throw e;
             }
         }
 
@@ -329,6 +338,24 @@ namespace PerCederberg.Grammatica.Parser {
             return bestMatch;
         }
 
+        /**
+         * Returns a string representation of this object. The returned
+         * string will contain the details of all the token patterns 
+         * contained in this tokenizer.
+         * 
+         * @return a detailed string representation 
+         */
+        public override string ToString() {
+            StringBuilder  buffer = new StringBuilder();
+            TokenMatcher   m;
+
+            for (int i = 0; i < matchers.Count; i++) {
+                m = (TokenMatcher) matchers[i];
+                buffer.Append(m);
+                buffer.Append('\n');
+            }
+            return buffer.ToString();
+        }
     }
 
 
@@ -400,6 +427,15 @@ namespace PerCederberg.Grammatica.Parser {
          *         false otherwise
          */
         public abstract bool MatchFrom(string str, int pos);
+
+        /**
+         * Returns a string representation of this token matcher.
+         * 
+         * @return a detailed string representation of this matcher
+         */
+        public override string ToString() {
+            return pattern.ToString() + "\n";
+        }
     }
     
 
@@ -588,5 +624,14 @@ namespace PerCederberg.Grammatica.Parser {
             matcher = regExp.Matcher(str);
             return matcher.MatchFrom(pos);
         }
+
+        /**
+         * Returns a string representation of this token matcher.
+         * 
+         * @return a detailed string representation of this matcher
+         */
+        public override string ToString() {
+            return base.ToString() + regExp.ToString();
+        }
     }
 }
