@@ -40,9 +40,9 @@ import java.io.Reader;
 import java.util.HashMap;
 import java.util.Vector;
 
-import net.percederberg.grammatica.parser.ParseException;
 import net.percederberg.grammatica.parser.Parser;
 import net.percederberg.grammatica.parser.ParserCreationException;
+import net.percederberg.grammatica.parser.ParserLogException;
 import net.percederberg.grammatica.parser.ProductionPattern;
 import net.percederberg.grammatica.parser.RecursiveDescentParser;
 import net.percederberg.grammatica.parser.TokenPattern;
@@ -108,6 +108,12 @@ public class Grammar extends Object {
      */
     private Vector tokens = new Vector();
 
+    /**
+     * The token id map. This is a map from the token pattern id to 
+     * the token pattern.
+     */
+    private HashMap tokenIds = new HashMap();
+
     /**
      * The token name map. This is map from the token pattern name to 
      * the token pattern. 
@@ -125,6 +131,12 @@ public class Grammar extends Object {
      */
     private Vector productions = new Vector();
 
+    /**
+     * The production id map. This is a map from the production 
+     * pattern id to the production pattern.
+     */
+    private HashMap productionIds = new HashMap();
+
     /**
      * The production name map. This is map from the production 
      * pattern name to the production pattern. 
@@ -143,12 +155,12 @@ public class Grammar extends Object {
      *
      * @throws FileNotFoundException if the grammar file could not be
      *             found
-     * @throws ParseException if the grammar file couldn't be parsed
-     *             correctly
+     * @throws ParserLogException if the grammar file couldn't be 
+     *             parsed correctly
      * @throws GrammarException if the grammar wasn't valid
      */
     public Grammar(File file) throws FileNotFoundException, 
-        ParseException, GrammarException {
+        ParserLogException, GrammarException {
 
         GrammarParser       parser;
         FirstPassAnalyzer   first = new FirstPassAnalyzer(this);
@@ -306,13 +318,24 @@ public class Grammar extends Object {
     }
 
     /**
-     * Returns a specific token pattern.
+     * Returns a token pattern identified by its id.
+     * 
+     * @param id             the pattern id
+     * 
+     * @return the token pattern, or null
+     */
+    public TokenPattern getTokenPatternById(int id) {
+        return (TokenPattern) tokenIds.get(new Integer(id));
+    }
+
+    /**
+     * Returns a token pattern identified by its name.
      * 
      * @param name           the pattern name
      * 
-     * @return the token pattern
+     * @return the token pattern, or null
      */
-    public TokenPattern getTokenPattern(String name) {
+    public TokenPattern getTokenPatternByName(String name) {
         return (TokenPattern) tokenNames.get(name);
     }
 
@@ -322,7 +345,7 @@ public class Grammar extends Object {
      * 
      * @param image          the pattern string
      * 
-     * @return the token pattern
+     * @return the token pattern, or null
      */
     TokenPattern getTokenPatternByImage(String image) {
         return (TokenPattern) tokenPatterns.get(image);
@@ -349,13 +372,24 @@ public class Grammar extends Object {
     }
     
     /**
-     * Returns a specific production pattern.
+     * Returns a production pattern identified by its id.
+     * 
+     * @param id             the pattern id
+     * 
+     * @return the production pattern, or null
+     */
+    public ProductionPattern getProductionPatternById(int id) {
+        return (ProductionPattern) productionIds.get(new Integer(id));
+    }
+
+    /**
+     * Returns a production pattern identified by its name.
      * 
      * @param name           the pattern name
      * 
-     * @return the production pattern
+     * @return the production pattern, or null
      */
-    public ProductionPattern getProductionPattern(String name) {
+    public ProductionPattern getProductionPatternByName(String name) {
         return (ProductionPattern) productionNames.get(name);
     }
 
@@ -378,6 +412,7 @@ public class Grammar extends Object {
      */
     void addToken(TokenPattern token, int start, int end) {
         tokens.add(token);
+        tokenIds.put(new Integer(token.getId()), token);
         tokenNames.put(token.getName(), token);
         if (token.getType() == TokenPattern.STRING_TYPE) {
             tokenPatterns.put(token.getPattern(), token);
@@ -394,6 +429,7 @@ public class Grammar extends Object {
      */
     void addProduction(ProductionPattern production, int start, int end) {
         productions.add(production);
+        productionIds.put(new Integer(production.getId()), production);
         productionNames.put(production.getName(), production);
         lines.put(production.getName(), new LineRange(start, end));
     }
