@@ -44,6 +44,7 @@ import net.percederberg.grammatica.output.JavaParserGenerator;
 import net.percederberg.grammatica.parser.ParseException;
 import net.percederberg.grammatica.parser.Parser;
 import net.percederberg.grammatica.parser.ParserCreationException;
+import net.percederberg.grammatica.parser.ParserLogException;
 import net.percederberg.grammatica.parser.Token;
 import net.percederberg.grammatica.parser.Tokenizer;
 
@@ -61,9 +62,16 @@ public class Grammatica extends Object {
      * The command-line help output.
      */
     private static final String COMMAND_HELP =
+        "Generates source code for a C# or Java parser from a grammar\n" +
+        "file. This program comes with ABSOLUTELY NO WARRANTY; for\n" +
+        "details see the LICENSE.txt file.\n" +
+        "\n" +
         "Syntax: Grammatica <grammarfile> <action> [<options>]\n" +
         "\n" +
         "Actions:\n" +
+        "  --debug\n" +
+        "      Debugs the grammar by validating it and printing the\n" +
+        "      internal representation.\n" +
         "  --tokenize <file>\n" +
         "      Debugs the grammar by using it to tokenize the specified\n" +
         "      file. No code has to be generated for this.\n" +
@@ -119,8 +127,8 @@ public class Grammatica extends Object {
             printHelp(null);
             System.exit(1);
         }
-        if (args.length < 3 || args.length > 7) {
-            printHelp("Wrong number of arguments");
+        if (args.length < 2) {
+            printHelp("Missing grammar file and/or action");
             System.exit(1);
         }
 
@@ -130,7 +138,7 @@ public class Grammatica extends Object {
         } catch (FileNotFoundException e) {
             printError(args[0], e);
             System.exit(1);
-        } catch (ParseException e) {
+        } catch (ParserLogException e) {
             printError(args[0], e);
             System.exit(1);
         } catch (GrammarException e) {
@@ -145,7 +153,12 @@ public class Grammatica extends Object {
 
         // Check action parameter
         try {
-            if (args[1].equals("--tokenize")) {
+            if (args[1].equals("--debug")) {
+                debug(grammar);
+            } else if (args.length < 3) {
+                printHelp("missing action file parameter");
+                System.exit(1);
+            } else if (args[1].equals("--tokenize")) {
                 tokenize(grammar, new File(args[2]));
             } else if (args[1].equals("--parse")) {
                 parse(grammar, new File(args[2]));
@@ -218,12 +231,6 @@ public class Grammatica extends Object {
         StringBuffer  buffer = new StringBuffer();
         String        line;
 
-        // Check for internal errror
-        if (e.getErrorType() == ParseException.INTERNAL_ERROR) {
-            printInternalError(e);
-            return;
-        }
-
         // Handle normal parse error
         buffer.append("Error: in ");
         buffer.append(file);
@@ -249,6 +256,18 @@ public class Grammatica extends Object {
         System.err.println(buffer.toString());
     }
     
+    /**
+     * Prints a list of parse error messages.
+     * 
+     * @param file           the input file name
+     * @param e              the parser log exception
+     */
+    private static void printError(String file, ParserLogException e) {
+        for (int i = 0; i < e.getErrorCount(); i++) {
+            printError(file, e.getError(i));
+        }
+    }
+
     /**
      * Prints a grammar error message.
      * 
@@ -345,6 +364,7 @@ public class Grammatica extends Object {
     private static String readLines(String file, int start, int end) {
         BufferedReader  input;
         StringBuffer    buffer = new StringBuffer();
+        String          str;
 
         // Check invalid line number
         if (start < 1 || end < start) {
@@ -355,10 +375,12 @@ public class Grammatica extends Object {
         try {
             input = new BufferedReader(new FileReader(file));
             for (int i = 0; i < end; i++) {
-                if (i + 1 < start) {
-                    input.readLine();
-                } else {
-                    buffer.append(input.readLine());
+                str = input.readLine();
+                if (str == null) {
+                    input.close();
+                    return null;
+                } else if (start <= i + 1) {
+                    buffer.append(str);
                     buffer.append("\n");
                 }
             }
@@ -370,6 +392,35 @@ public class Grammatica extends Object {
         return buffer.toString();
     }
 
+    /**
+     * Debugs a grammar by printing the internal representation.
+     * 
+     * @param grammar        the grammar to use
+     */
+    private static void debug(Grammar grammar) {
+        Tokenizer  tokenizer = null;
+        Parser     parser = null;
+        
+        // Create tokenizer and parser
+        try {
+            tokenizer = grammar.createTokenizer(null);
+            parser = grammar.createParser(tokenizer);
+        } catch (GrammarException e) {
+            printInternalError(e);
+            System.exit(2);
+        }
+
+        // Print tokenizer and parser
+        System.out.println("Contents of " + grammar.getFileName() + ":");
+        System.out.println();
+        System.out.println("Token Declarations:");
+        System.out.println("-------------------");
+        System.out.print(tokenizer);
+        System.out.println("Production Declarations:");
+        System.out.println("------------------------");
+        System.out.print(parser);
+    }
+
     /**
      * Tokenizes the specified file with the token patterns from the
      * grammar.
@@ -424,7 +475,7 @@ public class Grammatica extends Object {
         } catch (ParserCreationException e) {
             printInternalError(e);
             System.exit(2);
-        } catch (ParseException e) {
+        } catch (ParserLogException e) {
             printError(file.toString(), e);
             System.exit(1);
         }
