@@ -84,6 +84,12 @@ public class CSharpParserGenerator extends ParserGenerator {
 
         // Set base name
         str = getGrammar().getFileName();
+        if (str.indexOf('/') >= 0) {
+            str = str.substring(str.lastIndexOf('/') + 1);
+        }
+        if (str.indexOf('\\') >= 0) {
+            str = str.substring(str.lastIndexOf('\\') + 1);
+        }
         if (str.indexOf('.') > 0) {
             str = str.substring(0, str.indexOf('.'));
         }
