@@ -44,15 +44,10 @@ import java.util.Iterator;
  * that is has to consider.
  *
  * @author   Per Cederberg, <per at percederberg dot net>
- * @version  1.0
+ * @version  1.1
  */
 public class RecursiveDescentParser extends Parser {
 
-    /**
-     * The parser initialization flag.
-     */
-    private boolean initialized = false;
-
     /**
      * The map of pattern look-ahead sets. The map is indexed by the
      * production pattern object. 
@@ -110,7 +105,6 @@ public class RecursiveDescentParser extends Parser {
 
         // Add pattern 
         super.addPattern(pattern);
-        initialized = false;
     }
 
     /**
@@ -123,10 +117,11 @@ public class RecursiveDescentParser extends Parser {
      *             initialized correctly 
      */
     public void prepare() throws ParserCreationException {
-        Iterator           iter;
+        Iterator  iter;
 
         // Performs production pattern checks
         super.prepare();
+        setInitialized(false);
 
         // Calculate production look-ahead sets
         iter = getPatterns().iterator();
@@ -135,7 +130,7 @@ public class RecursiveDescentParser extends Parser {
         }
 
         // Set initialized flag
-        initialized = true;
+        setInitialized(true);
     }
 
     /**
@@ -143,21 +138,14 @@ public class RecursiveDescentParser extends Parser {
      * 
      * @return the parse tree
      * 
-     * @throws ParserCreationException if the parser couldn't be
-     *             initialized correctly
      * @throws ParseException if the input couldn't be parsed 
      *             correctly
-     * 
-     * @see #prepare
      */
-    public Node parse() throws ParserCreationException, ParseException {
+    protected Node parseStart() throws ParseException {
         Token      token;
         Node       node;
         ArrayList  list;
         
-        if (!initialized) {
-            prepare();
-        }
         node = parsePattern(getStartPattern());
         token = peekToken(0); 
         if (token != null) {
@@ -223,7 +211,13 @@ public class RecursiveDescentParser extends Parser {
         node = new Production(alt.getPattern());
         enterNode(node);
         for (int i = 0; i < alt.getElementCount(); i++) {
-            parseElement(node, alt.getElement(i));
+            try {
+                parseElement(node, alt.getElement(i));
+            } catch (ParseException e) {
+                addError(e, true);
+                nextToken();
+                i--;
+            }
         }
         return exitNode(node);
     }
@@ -270,13 +264,8 @@ public class RecursiveDescentParser extends Parser {
      * 
      * @return true if the next tokens match, or
      *         false otherwise
-     * 
-     * @throws ParseException if a token couldn't be read or parsed 
-     *             correctly
      */
-    private boolean isNext(ProductionPattern pattern) 
-        throws ParseException {
-
+    private boolean isNext(ProductionPattern pattern) {
         LookAheadSet  set = pattern.getLookAhead();
 
         if (set == null) {
@@ -295,13 +284,8 @@ public class RecursiveDescentParser extends Parser {
      *  
      * @return true if the next tokens match, or
      *         false otherwise
-     * 
-     * @throws ParseException if a token couldn't be read or parsed 
-     *             correctly
      */
-    private boolean isNext(ProductionPatternAlternative alt)
-        throws ParseException {
-
+    private boolean isNext(ProductionPatternAlternative alt) {
         LookAheadSet  set = alt.getLookAhead();
 
         if (set == null) {
@@ -321,13 +305,8 @@ public class RecursiveDescentParser extends Parser {
      *  
      * @return true if the next tokens match, or
      *         false otherwise
-     * 
-     * @throws ParseException if a token couldn't be read or parsed 
-     *             correctly
      */
-    private boolean isNext(ProductionPatternElement elem) 
-        throws ParseException {
-
+    private boolean isNext(ProductionPatternElement elem) {
         LookAheadSet  set = elem.getLookAhead();
         
         if (set != null) {
@@ -354,8 +333,9 @@ public class RecursiveDescentParser extends Parser {
 
         ProductionPatternAlternative  alt;
         LookAheadSet                  result;
-        LookAheadSet                  alternatives[];
+        LookAheadSet[]                alternatives;
         LookAheadSet                  conflicts;
+        LookAheadSet                  previous = new LookAheadSet(0);
         int                           length = 1;
         int                           i;
         CallStack                     stack = new CallStack();
@@ -365,9 +345,9 @@ public class RecursiveDescentParser extends Parser {
         result = new LookAheadSet(1);
         alternatives = new LookAheadSet[pattern.getAlternativeCount()];
         for (i = 0; i < pattern.getAlternativeCount(); i++) {
-            alternatives[i] = findLookAhead(pattern.getAlternative(i), 
-                                            1,
-                                            stack);
+            alt = pattern.getAlternative(i);
+            alternatives[i] = findLookAhead(alt, 1, 0, stack, null);
+            alt.setLookAhead(alternatives[i]);
             result.addAll(alternatives[i]);
         }
         if (pattern.getLookAhead() == null) {
@@ -380,12 +360,15 @@ public class RecursiveDescentParser extends Parser {
             length++;
             stack.clear();
             stack.push(pattern.getName(), length);
+            conflicts.addAll(previous);
             for (i = 0; i < pattern.getAlternativeCount(); i++) {
                 alt = pattern.getAlternative(i);
                 if (alternatives[i].intersects(conflicts)) {
                     alternatives[i] = findLookAhead(alt, 
                                                     length,
-                                                    stack);
+                                                    0,
+                                                    stack,
+                                                    conflicts);
                     alt.setLookAhead(alternatives[i]);
                 }
                 if (alternatives[i].intersects(conflicts)) {
@@ -399,6 +382,7 @@ public class RecursiveDescentParser extends Parser {
                     }
                 }
             }
+            previous = conflicts;
             conflicts = findConflicts(pattern, length);
         }
         
@@ -429,6 +413,7 @@ public class RecursiveDescentParser extends Parser {
         LookAheadSet              first;
         LookAheadSet              follow;
         LookAheadSet              conflicts;
+        LookAheadSet              previous = new LookAheadSet(0);
         String                    location;
         int                       length = 1;
 
@@ -446,8 +431,8 @@ public class RecursiveDescentParser extends Parser {
         }
 
         // Calculate simple look-aheads
-        first = findLookAhead(elem, 1, new CallStack());
-        follow = findLookAhead(alt, 1, pos + 1, new CallStack()); 
+        first = findLookAhead(elem, 1, new CallStack(), null);
+        follow = findLookAhead(alt, 1, pos + 1, new CallStack(), null); 
         
         // Resolve conflicts
         location = "at position " + (pos + 1);
@@ -457,14 +442,23 @@ public class RecursiveDescentParser extends Parser {
                                   follow);
         while (conflicts.size() > 0) {
             length++;
-            first = findLookAhead(elem, length, new CallStack());
-            follow = findLookAhead(alt, length, pos + 1, new CallStack());
+            conflicts.addAll(previous);
+            first = findLookAhead(elem, 
+                                  length, 
+                                  new CallStack(), 
+                                  conflicts);
+            follow = findLookAhead(alt, 
+                                   length, 
+                                   pos + 1, 
+                                   new CallStack(), 
+                                   conflicts);
             first = first.createCombination(follow);
             elem.setLookAhead(first);
             if (first.intersects(conflicts)) {
                 first = first.createIntersection(conflicts);
                 throwAmbiguityException(pattern.getName(), location, first);
             }
+            previous = conflicts;
             conflicts = findConflicts(pattern.getName(), 
                                       location, 
                                       first, 
@@ -477,13 +471,14 @@ public class RecursiveDescentParser extends Parser {
 
     /**
      * Finds the look-ahead set for a production pattern. The maximum 
-     * look-ahead length must be specified. This method will cache 
-     * some results in the pattern look-ahead wrapper for improved 
-     * performance.
+     * look-ahead length must be specified. It is also possible to 
+     * specify a look-ahead set filter, which will make sure that 
+     * unnecessary token sequences will be avoided.
      * 
      * @param pattern        the production pattern
      * @param length         the maximum look-ahead length
      * @param stack          the call stack used for loop detection
+     * @param filter         the look-ahead set filter
      * 
      * @return the look-ahead set for the production pattern
      * 
@@ -492,7 +487,8 @@ public class RecursiveDescentParser extends Parser {
      */
     private LookAheadSet findLookAhead(ProductionPattern pattern, 
                                        int length,
-                                       CallStack stack) 
+                                       CallStack stack,
+                                       LookAheadSet filter) 
         throws ParserCreationException {
 
         LookAheadSet  result;
@@ -506,82 +502,34 @@ public class RecursiveDescentParser extends Parser {
                 (String) null);
         }
 
-        // Check for cached result
-        result = pattern.getLookAhead();
-        if (result != null) {
-            if (length == result.getMaxLength()) {
-                return result;
-            } else if (length < result.getMaxLength()) {
-                return new LookAheadSet(length, result);
-            }
-        }
-
         // Find pattern look-ahead
         stack.push(pattern.getName(), length);
         result = new LookAheadSet(length);
         for (int i = 0; i < pattern.getAlternativeCount(); i++) {
-            temp = findLookAhead(pattern.getAlternative(i), length, stack);
+            temp = findLookAhead(pattern.getAlternative(i), 
+                                 length,
+                                 0,
+                                 stack,
+                                 filter);
             result.addAll(temp);
         }
-        if (pattern.getLookAhead() == null) {
-            pattern.setLookAhead(result);
-        }
         stack.pop();
 
         return result;
     }
 
-    /**
-     * Finds the look-ahead set for a production pattern alternative. 
-     * The maximum look-ahead length must be specified. This method 
-     * will cache some results in the pattern look-ahead wrapper for 
-     * improved performance.
-     * 
-     * @param alt            the production pattern alternative
-     * @param length         the maximum look-ahead length
-     * @param stack          the call stack used for loop detection
-     * 
-     * @return the look-ahead set for the pattern alternative
-     * 
-     * @throws ParserCreationException if an infinite loop was found
-     *             in the grammar
-     */
-    private LookAheadSet findLookAhead(ProductionPatternAlternative alt, 
-                                       int length,
-                                       CallStack stack) 
-        throws ParserCreationException {
-
-        LookAheadSet  result;
-                                
-        // Check for cached result
-        result = alt.getLookAhead();
-        if (result != null) {
-            if (length == result.getMaxLength()) {
-                return result;
-            } else if (length < result.getMaxLength()) {
-                return new LookAheadSet(length, result); 
-            }
-        }
-
-        // Find rule look-ahead
-        result = findLookAhead(alt, length, 0, stack);
-        if (alt.getLookAhead() == null) {
-            alt.setLookAhead(result);
-        }
-
-        return result;
-    }
-
-
     /**
      * Finds the look-ahead set for a production pattern alternative. 
      * The pattern position and maximum look-ahead length must be 
-     * specified.
+     * specified. It is also possible to specify a look-ahead set 
+     * filter, which will make sure that unnecessary token sequences 
+     * will be avoided.
      * 
      * @param alt            the production pattern alternative
      * @param length         the maximum look-ahead length
      * @param pos            the pattern element position
      * @param stack          the call stack used for loop detection
+     * @param filter         the look-ahead set filter
      * 
      * @return the look-ahead set for the pattern alternative
      * 
@@ -591,11 +539,13 @@ public class RecursiveDescentParser extends Parser {
     private LookAheadSet findLookAhead(ProductionPatternAlternative alt,
                                        int length,
                                        int pos,
-                                       CallStack stack) 
+                                       CallStack stack,
+                                       LookAheadSet filter) 
         throws ParserCreationException {
 
         LookAheadSet  first;
         LookAheadSet  follow;
+        LookAheadSet  overlaps;
 
         // Check trivial cases
         if (length <= 0 || pos >= alt.getElementCount()) {
@@ -603,19 +553,28 @@ public class RecursiveDescentParser extends Parser {
         }
 
         // Find look-ahead for this element
-        first = findLookAhead(alt.getElement(pos), length, stack);
+        first = findLookAhead(alt.getElement(pos), length, stack, filter);
         if (alt.getElement(pos).getMinCount() == 0) {
             first.addEmpty();
         }
         
         // Find remaining look-ahead
-        length -= first.getMinLength();
-        if (length <= 0) {
-            return first;
-        } else {
-            follow = findLookAhead(alt, length, pos + 1, stack); 
-            return first.createCombination(follow); 
+        if (filter == null) {
+            length -= first.getMinLength();
+            if (length > 0) {
+                follow = findLookAhead(alt, length, pos + 1, stack, null); 
+                first = first.createCombination(follow);
+            }
+        } else if (filter.isOverlap(first)) {
+            overlaps = first.createOverlaps(filter);
+            length -= overlaps.getMinLength();
+            filter = filter.createFilter(overlaps);
+            follow = findLookAhead(alt, length, pos + 1, stack, filter); 
+            first.removeAll(overlaps);
+            first.addAll(overlaps.createCombination(follow));
         }
+
+        return first;
     }
 
     /**
@@ -623,11 +582,14 @@ public class RecursiveDescentParser extends Parser {
      * maximum look-ahead length must be specified. This method takes
      * the element repeats into consideration when creating the 
      * look-ahead set, but does NOT include an empty sequence even if
-     * the minimum count is zero (0). 
+     * the minimum count is zero (0). It is also possible to specify a
+     * look-ahead set filter, which will make sure that unnecessary 
+     * token sequences will be avoided.
      * 
      * @param elem           the production pattern element
      * @param length         the maximum look-ahead length
      * @param stack          the call stack used for loop detection
+     * @param filter         the look-ahead set filter
      * 
      * @return the look-ahead set for the pattern element
      * 
@@ -636,38 +598,86 @@ public class RecursiveDescentParser extends Parser {
      */
     private LookAheadSet findLookAhead(ProductionPatternElement elem,
                                        int length,
-                                       CallStack stack) 
+                                       CallStack stack,
+                                       LookAheadSet filter) 
         throws ParserCreationException {
 
         ProductionPattern  pattern;
         LookAheadSet       result;
-        LookAheadSet       set;
+        LookAheadSet       first;
+        LookAheadSet       follow;
         int                max;
 
-        // Find look-ahead for this element
-        if (elem.isToken()) {
-            set = new LookAheadSet(length);
-            set.add(elem.getId());
-        } else {
-            pattern = getPattern(elem.getId());
-            set = findLookAhead(pattern, length, stack);
-            if (stack.contains(pattern.getName())) {
-                set = set.createRepetitive();
-            }
+        // Find initial element look-ahead
+        first = findLookAhead(elem, length, 0, stack, filter);
+        result = new LookAheadSet(length);
+        result.addAll(first);
+        if (filter == null || !filter.isOverlap(result)) {
+            return result;
         }
 
         // Handle element repetitions
-        result = new LookAheadSet(length);
-        result.addAll(set);
         if (elem.getMaxCount() == Integer.MAX_VALUE) {
-            set = set.createRepetitive();
+            first = first.createRepetitive();
         }
         max = Math.min(length, elem.getMaxCount());
         for (int i = 1; i < max; i++) {
-            set = set.createCombination(set);
-            result.addAll(set);
+            first = first.createOverlaps(filter);
+            if (first.size() <= 0 || first.getMinLength() >= length) {
+                break;
+            }
+            follow = findLookAhead(elem, 
+                                   length, 
+                                   0, 
+                                   stack, 
+                                   filter.createFilter(first));
+            first = first.createCombination(follow);
+            result.addAll(first);
         }
-        
+
+        return result;
+    }
+
+    /**
+     * Finds the look-ahead set for a production pattern element. The
+     * maximum look-ahead length must be specified. This method does
+     * NOT take the element repeat into consideration when creating 
+     * the look-ahead set. It is also possible to specify a look-ahead
+     * set filter, which will make sure that unnecessary token 
+     * sequences will be avoided.
+     * 
+     * @param elem           the production pattern element
+     * @param length         the maximum look-ahead length
+     * @param dummy          a parameter to distinguish the method
+     * @param stack          the call stack used for loop detection
+     * @param filter         the look-ahead set filter
+     * 
+     * @return the look-ahead set for the pattern element
+     * 
+     * @throws ParserCreationException if an infinite loop was found
+     *             in the grammar
+     */
+    private LookAheadSet findLookAhead(ProductionPatternElement elem,
+                                       int length,
+                                       int dummy,
+                                       CallStack stack,
+                                       LookAheadSet filter) 
+        throws ParserCreationException {
+
+        LookAheadSet       result;
+        ProductionPattern  pattern;
+
+        if (elem.isToken()) {
+            result = new LookAheadSet(length);
+            result.add(elem.getId());
+        } else {
+            pattern = getPattern(elem.getId());
+            result = findLookAhead(pattern, length, stack, filter);
+            if (stack.contains(pattern.getName())) {
+                result = result.createRepetitive();
+            }
+        }
+
         return result;
     }
 
