@@ -50,7 +50,7 @@ import net.percederberg.grammatica.parser.re.RegExpException;
  * don't match any of the token patterns, a parse exception is thrown. 
  *
  * @author   Per Cederberg, <per at percederberg dot net>
- * @version  1.0
+ * @version  1.1
  */
 public class Tokenizer {
     
@@ -228,9 +228,10 @@ public class Tokenizer {
      *             parsed correctly
      */
     private Token nextToken() throws ParseException {
-        TokenMatcher  m;
-        Token         token;
-        String        str;
+        TokenMatcher    m;
+        Token           token;
+        String          str;
+        ParseException  e;
 
         // Find longest matching string 
         do {
@@ -252,11 +253,19 @@ public class Tokenizer {
         } else if (position >= buffer.length()) {
             return null;
         } else {
-            throw new ParseException(
+            e = new ParseException(
                 ParseException.UNEXPECTED_CHAR_ERROR,
                 String.valueOf(buffer.charAt(position)),
                 line,
-                column);
+                column); 
+            if (buffer.charAt(position) == '\n') {
+                line++;
+                column = 1;
+            } else {
+                column++;
+            }
+            position++;
+            throw e;
         }
     }
     
@@ -335,8 +344,27 @@ public class Tokenizer {
         }
         return bestMatch;
     }
+
+    /**
+     * Returns a string representation of this object. The returned
+     * string will contain the details of all the token patterns 
+     * contained in this tokenizer.
+     * 
+     * @return a detailed string representation 
+     */
+    public String toString() {
+        StringBuffer  buffer = new StringBuffer();
+        TokenMatcher  m;
+
+        for (int i = 0; i < matchers.size(); i++) {
+            m = (TokenMatcher) matchers.get(i);
+            buffer.append(m);
+            buffer.append('\n');
+        }
+        return buffer.toString();
+    }
     
-    
+
     /**
      * A token pattern matcher. This class is the base class for the
      * two types of token matchers that exist. The token matcher 
@@ -404,6 +432,15 @@ public class Tokenizer {
          *         false otherwise
          */
         public abstract boolean matchFrom(int pos);
+        
+        /**
+         * Returns a string representation of this token matcher.
+         * 
+         * @return a detailed string representation of this matcher
+         */
+        public String toString() {
+            return pattern.toString() + "\n";
+        }
     }
     
 
@@ -515,6 +552,11 @@ public class Tokenizer {
      */
     private class RegExpTokenMatcher extends TokenMatcher {
 
+        /**
+         * The regular expression to use.
+         */
+        private RegExp regExp;
+
         /**
          * The regular expression matcher to use.
          */
@@ -532,7 +574,7 @@ public class Tokenizer {
             throws RegExpException {
 
             super(pattern);
-            RegExp regExp = new RegExp(pattern.getPattern());
+            regExp = new RegExp(pattern.getPattern());
             matcher = regExp.matcher(buffer);
         }
 
@@ -584,5 +626,14 @@ public class Tokenizer {
         public boolean matchFrom(int pos) {
             return matcher.matchFrom(pos);
         }
+        
+        /**
+         * Returns a string representation of this token matcher.
+         * 
+         * @return a detailed string representation of this matcher
+         */
+        public String toString() {
+            return super.toString() + regExp.toString();
+        }
     }
 }
