@@ -126,9 +126,8 @@ public class RegExp {
         StringWriter  str;
         
         str = new StringWriter();
-        str.write("Regular Expression\n");
-        str.write("  Pattern: " + pattern + "\n");
-        str.write("  Compiled:\n");
+        str.write("  Regexp Pattern: " + pattern + "\n");
+        str.write("  Regexp Compiled:\n");
         element.printTo(new PrintWriter(str), "    ");
         return str.toString();
     }
