@@ -42,7 +42,7 @@ import java.util.BitSet;
  * repetition count.
  *
  * @author   Per Cederberg, <per at percederberg dot net>
- * @version  1.0
+ * @version  1.1
  */
 class RepeatElement extends Element {
 
@@ -143,11 +143,15 @@ class RepeatElement extends Element {
      *         -1 if no match was found
      */
     public int match(Matcher m, CharBuffer str, int start, int skip) {
+        if (skip == 0) {
+            matchStart = -1;
+            matches = null;
+        }
         switch (type) {
         case GREEDY:
             return matchGreedy(m, str, start, skip);
         case RELUCTANT:
-            return matchReluctant(m, str, start, 1, skip, 0);
+            return matchReluctant(m, str, start, skip);
         case POSSESSIVE:
             if (skip == 0) {
                 return matchPossessive(m, str, start, 0);
@@ -200,77 +204,40 @@ class RepeatElement extends Element {
     }
 
     /**
-     * Returns the length of the minimum number of elements matching 
-     * the string starting at the specified position. The number of 
-     * matches to skip here and in the rest of the matches must also 
-     * be specified.
+     * Returns the length of the shortest possible matchine string
+     * starting at the specified position. The number of matches to 
+     * skip can also be specified.
      *
      * @param m              the matcher being used 
      * @param str            the string to match
      * @param start          the starting position
-     * @param count          the number of matched elements
-     * @param skipHere       the number of skips at this position
-     * @param skipNext       the number of skips at later positions
+     * @param skip           the number of matches to skip
      * 
-     * @return the length of the longest matching string, or
+     * @return the length of the shortest matching string, or
      *         -1 if no match was found
      */
     private int matchReluctant(Matcher m, 
                                CharBuffer str, 
                                int start,
-                               int count,
-                               int skipHere, 
-                               int skipNext) {
-
-        int  length;
-        int  nextLength;
-
-        // Check for empty match
-        if (count == 1 && min == 0 && skipNext == 0) {
-            if (skipHere == 0) {
-                return 0;
-            } else {
-                skipHere--;
-            }
+                               int skip) {
+    
+        // Find all matches
+        if (matchStart != start) {
+            matchStart = start;
+            matches = new BitSet();
+            findMatches(m, str, start, 0, 0, 0);
         }
 
-        // Check element match
-        length = elem.match(m, str, start, skipHere);
-
-        // Check for possible match here
-        if (length >= 0) {
-            if (min <= count) {
-                if (skipNext == 0) {
-                    return length;
-                } else {
-                    skipNext--;
+        // Find first non-skipped match
+        for (int i = 0; i <= matches.size(); i++) {
+            if (matches.get(i)) {
+                if (skip == 0) {
+                    return i;
                 }
-            }
-            if (length == 0) {
-                return -1;
-            }
-            nextLength = matchReluctant(m, 
-                                        str, 
-                                        start + length, 
-                                        count + 1, 
-                                        skipNext,
-                                        0);
-            if (nextLength >= 0) {
-                return length + nextLength;
+                skip--;
             }
         }
-
-        // Check for subsequent matches
-        if (skipHere == 0) {
-            return -1;
-        } else {
-            return matchReluctant(m, 
-                                  str, 
-                                  start,
-                                  count, 
-                                  skipHere - 1, 
-                                  skipNext + 1);
-        } 
+        return -1;
     }
 
     /**
@@ -287,9 +254,9 @@ class RepeatElement extends Element {
      *         -1 if no match was found
      */
     private int matchPossessive(Matcher m, 
-                               CharBuffer str, 
-                               int start,
-                               int count) {
+                                CharBuffer str, 
+                                int start,
+                                int count) {
                                    
         int  length = 0;
         int  subLength = 1;
