@@ -59,8 +59,8 @@ namespace PerCederberg.Grammatica.Test {
                 ValidateTree(parser.Parse(), output);
             } catch (ParserCreationException e) {
                 Fail(e.Message);
-            } catch (ParseException e) {
-                Fail(e.Message);
+            } catch (ParserLogException e) {
+                Fail(e.GetError(0).GetMessage());
             }
         }
         
@@ -84,10 +84,13 @@ namespace PerCederberg.Grammatica.Test {
                 Fail("parsing succeeded");
             } catch (ParserCreationException e) {
                 Fail(e.Message);
-            } catch (ParseException e) {
-                AssertEquals("error type", type, e.GetErrorType());
-                AssertEquals("line number", line, e.GetLine());
-                AssertEquals("column number", column, e.GetColumn());
+            } catch (ParserLogException e) {
+                ParseException  p = e.GetError(0);  
+            
+                AssertEquals("error count", 1, e.GetErrorCount());
+                AssertEquals("error type", type, p.GetErrorType());
+                AssertEquals("line number", line, p.GetLine());
+                AssertEquals("column number", column, p.GetColumn());
             }
         }
     
